<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Sascha Nitsch (grumpydeveloper)
 **/

namespace Federator\Connector;

/**
 * Connector to Mastodon.social
 */
class Mastodon implements Connector
{
    /**
     * config parameter
     *
     * @var array<string, mixed> $config
     */
    private $config;

    /**
     * main instance
     *
     * @var \Federator\Main $main
     */
    private $main;

    /**
     * service-URL
     *
     * @var string $service
     */
    private $service;

    /**
     * constructor
     *
     * @param \Federator\Main $main
     */
    public function __construct($main)
    {
        $config = parse_ini_file($_SERVER['DOCUMENT_ROOT'] . '../mastodon.ini', true);
        if ($config !== false) {
            $this->config = $config;
        }
        $this->service = $config['mastodon']['service-uri'];
        $this->main = $main;
    }
    
    /**
     * get the host this connector is dedicated to
     *
     * @return string
     */
    public function getHost() {
        return "mastodon.social";
    }

    /**
     * get posts by given user
     *
     * @param string $userId user id
     * @param string $min min date
     * @param string $max max date
     * @return \Federator\Data\ActivityPub\Common\APObject[]|false
     */
    public function getRemotePostsByUser($userId, $min = null, $max = null)
    {
        $remoteURL = $this->service . '/users/' . $userId . '/outbox';

        $items = [];

        do {
            if ($min !== '') {
                $remoteURL .= '&minTS=' . urlencode($min);
            }
            if ($max !== '') {
                $remoteURL .= '&maxTS=' . urlencode($max);
            }
            // Fetch the current page of items (first or subsequent pages)
            [$outboxResponse, $outboxInfo] = \Federator\Main::getFromRemote($remoteURL, ['Accept: application/activity+json']);
            
            if ($outboxInfo['http_code'] !== 200) {
                echo "aborting";
                return false;
            }

            $outbox = json_decode($outboxResponse, true);

            // Extract orderedItems from the current page
            if (isset($outbox['orderedItems'])) {
                $items = array_merge($items, $outbox['orderedItems']);
            }

            // Step 4: Use 'last' URL to determine pagination
            if (isset($outbox['last'])) {
                // The 'last' URL will usually have a query string that includes min_id for the next set of results
                $remoteURL = $outbox['last']; // Update to the last URL for the next page of items
            } else {
                break; // No more pages, exit pagination
            }

        } while (!empty($outbox['last'])); // Continue fetching until no 'last' URL


        $items = [];

        // Follow `first` page (or get orderedItems directly)
        if (isset($outbox['orderedItems'])) {
            $items = $outbox['orderedItems'];
        } elseif (isset($outbox['first'])) {
            $firstURL = is_array($outbox['first']) ? $outbox['first']['id'] : $outbox['first'];
            [$pageResponse, $pageInfo] = \Federator\Main::getFromRemote($firstURL, ['Accept: application/activity+json']);
            if ($pageInfo['http_code'] !== 200) {
                return false;
            }
            $page = json_decode($pageResponse, true);
            $items = $page['orderedItems'] ?? [];
        }

        // Convert to internal representation
        $posts = [];
        foreach ($items as $activity) {
            if (!isset($activity['type']) || $activity['type'] !== 'Create' || !isset($activity['object'])) {
                continue; // Skip non-Create activities
            }

            $obj = $activity['object'];
            $create = new \Federator\Data\ActivityPub\Common\Create();
            $create->setID($activity['id'])
                ->setPublished(strtotime($activity['published'] ?? $obj['published'] ?? 'now'))
                ->setAActor($activity['actor'])
                ->addTo("https://www.w3.org/ns/activitystreams#Public");

            // Handle main Note content
            if ($obj['type'] === 'Note') {
                $apNote = new \Federator\Data\ActivityPub\Common\Note();
                $apNote->setID($obj['id'])
                    ->setPublished(strtotime($obj['published'] ?? 'now'))
                    ->setContent($obj['content'] ?? '')
                    ->setAttributedTo($obj['attributedTo'] ?? $activity['actor'])
                    ->addTo("https://www.w3.org/ns/activitystreams#Public");

                // Handle attachments
                if (!empty($obj['attachment']) && is_array($obj['attachment'])) {
                    foreach ($obj['attachment'] as $media) {
                        if (!isset($media['type'], $media['url']))
                            continue;
                        $mediaObj = new \Federator\Data\ActivityPub\Common\APObject($media['type']);
                        $mediaObj->setURL($media['url']);
                        $apNote->addAttachment($mediaObj);
                    }
                }

                $create->setObject($apNote);
            }

            $posts[] = $create;
        }

        return $posts;
    }

    /**
     * Get the MIME type of a remote file by its URL.
     *
     * @param string $_url The URL of the remote file.
     * @return string|false The MIME type if found, or false on failure.
     */
    public function getRemoteMimeType($url)
    {
        $headers = get_headers($url, 1);
        return $headers['Content-Type'] ?? 'unknown';
    }

    /**
     * get statistics from remote system
     *
     * @return \Federator\Data\Stats|false
     */
    public function getRemoteStats()
    {
        $remoteURL = $this->service . '/api/stats';
        [$response, $info] = \Federator\Main::getFromRemote($remoteURL, []);
        if ($info['http_code'] != 200) {
            return false;
        }
        $r = json_decode($response, true);
        if ($r === false || $r === null || !is_array($r)) {
            return false;
        }
        $stats = new \Federator\Data\Stats();
        $stats->userCount = array_key_exists('userCount', $r) ? $r['userCount'] : 0;
        $stats->postCount = array_key_exists('pageCount', $r) ? $r['pageCount'] : 0;
        $stats->commentCount = array_key_exists('commentCount', $r) ? $r['commentCount'] : 0;
        return $stats;
    }

    /**
     * get remote user by given name
     *
     * @param string $_name user/profile name
     * @return \Federator\Data\User | false
     */
    public function getRemoteUserByName(string $_name)
    {
        // Validate username (Mastodon usernames can include @ and domain parts)
        if (preg_match("/^[a-zA-Z0-9_\-@.]+$/", $_name) !== 1) {
            return false;
        }

        // Mastodon lookup API endpoint
        $remoteURL = $this->service . '/api/v1/accounts/lookup?acct=' . urlencode($_name);

        // Set headers
        $headers = ['Accept: application/json'];

        // Fetch data from Mastodon instance
        [$response, $info] = \Federator\Main::getFromRemote($remoteURL, $headers);

        // Handle HTTP errors
        if ($info['http_code'] !== 200) {
            return false;
        }

        // Decode response
        $r = json_decode($response, true);
        if ($r === false || $r === null || !is_array($r)) {
            return false;
        }

        // Map response to User object
        $user = new \Federator\Data\User();
        $user->externalid = (string) $r['id']; // Mastodon uses numeric IDs
        $user->iconMediaType = 'image/png'; // Mastodon doesn't explicitly return this, assume PNG
        $user->iconURL = $r['avatar'] ?? null;
        $user->imageMediaType = 'image/png';
        $user->imageURL = $r['header'] ?? null;
        $user->name = $r['display_name'] ?: $r['username'];
        $user->summary = $r['note'];
        $user->type = 'Person'; // Mastodon profiles are ActivityPub "Person" objects
        $user->registered = strtotime($r['created_at']);

        return $user;
    }

    /**
     * get remote user by given session
     *
     * @param string $_session session id
     * @param string $_user user or profile name
     * @return \Federator\Data\User | false
     */
    public function getRemoteUserBySession(string $_session, string $_user)
    {
        // validate $_session and $user
        if (preg_match("/^[a-z0-9]{16}$/", $_session) != 1) {
            return false;
        }
        if (preg_match("/^[a-zA-Z0-9_\-]+$/", $_user) != 1) {
            return false;
        }
        $remoteURL = $this->service . '/api/users/permissions?profile=' . urlencode($_user);
        $headers = ['Cookie: session=' . $_session, 'Accept: application/json'];
        [$response, $info] = \Federator\Main::getFromRemote($remoteURL, $headers);

        if ($info['http_code'] != 200) {
            return false;
        }
        $r = json_decode($response, true);
        if ($r === false || !is_array($r) || !array_key_exists($_user, $r)) {
            return false;
        }
        $user = $this->getRemoteUserByName($_user);
        if ($user === false) {
            return false;
        }
        // extend with permissions
        $user->permissions = [];
        $user->session = $_session;
        foreach ($r[$_user] as $p) {
            $user->permissions[] = $p;
        }
        return $user;
    }
}

namespace Federator;

/**
 * Function to initialize plugin
 *
 * @param  \Federator\Main $main main instance
 * @return void
 */
function mastodon_load($main)
{
    $mast = new Connector\Mastodon($main);
    $main->addConnector($mast->getHost(), $mast);
}
